/* DUMP.C
 * Programmino di DUMP
 *
 * Nuova versione 2/7/93;
 * Questa versione aggiunge una serie di opzioni alla versione precedente:
 *	1) di default preferisce potenze di 2 come lunghezze linea (disabilitabile)
 *	2) Opzione di lunghezza record.
 *	3) Opzione di 'inizio dati'
 *	4) Opzione di 'interattivo'
 *	5) Opzione di 'numero record'
 *	6) Opzioni dimensione dati
 */

#include	<stdio.h>
#include	<stdlib.h>
#include	<fcntl.h>
#include	<string.h>
#include	<io.h>

#include	<conio.h>
#include	<bios.h>
#include	<dos.h>
#include	<mem.h>
#include	<ctype.h>

#include	"wins.h"

#define		MAXIN		8
#define		MAXFILS    14
#define		NORMAL		0x0F
#define		LOW			0x17
#define		REVERSE		0x70

void showframe (int argc,char **argv,int l);
void Options (char *hex,char *compare,int *irlen,int *rlen,char *align,
	int *reclen,int *ireclen,long *start,int *l,long *head,long maxl,
		int argc,char **argv);
void PrintHead (int argc,char **argv,int l);
void Warning (char *mess);
void Calculator (void);
long doopr (char opr,long acc,long op);
void PrintLines (int lines,int argc,char **argv,int l,int *fil,long start,
	long end,long head,int rlen,int reclen,char hex,char compare,char align);
int CalcRLen (int *l,int argc,char hex,int rlen,char align);
int Kind (char *value,int x,int y);
int	Boolean (char *value,int x,int y);
int Long (long *value,int x,int y);
int Int (int *value,int x,int y);
void initcursor(void);
void setcursor(unsigned int shape);
void changecursor (int insmode);
int strint (int *s,int c);
int GetInput (char *s,int max,int st,int y,int *term);

#define	ALTC	0x2e00
#define	ALTD    0x2000
#define	ALTH    0x2300
#define	ALTO    0x1800

int stdtetm [] = {
		ESC,CR,F9,F10,ARUP,ARDOW,0
	};

int calcterm [] = {
		ESC,CR,'+','-','*','/','|','&',ALTC,ALTD,ALTH,ALTO,'=',0
	};

void main (int argc,char **argv)
{
int		fil [MAXFILS];
long	start,end,head;
char    hex,compare,inter,align;
int 	rlen,irlen,reclen,ireclen;
int		i,l,key;
int		rcl,plen;
long	cl,maxl;

	hex = compare = inter = 0;
	rlen = reclen = 0;
	head = start = 0;
	end = 0x7fffffff;
	align = 0;	// default char display

	irlen = ireclen = 0;
	plen = -1;

	while (argc > 1 && (argv [argc - 1][0] == '-' ||
		argv [argc - 1][0] == '/'))
	{
		argc --;
		switch (argv [argc][1])
		{
			case 'h' :
			case 'H' :	/* Solo HEX (niente ASCII) */
				hex = 1;
				break;
			case 's' :
			case 'S' :	/* Start */
				start = strtol (argv [argc] + 2,(char **) 0,0);
				break;
			case 'e' :
			case 'E' :	/* End */
				end = strtol (argv [argc] + 2,(char **) 0,0);
				break;
			case 'c' :
			case 'C' :	/* Compare */
				compare = 1;
				break;
			case 't' :
			case 'T' :	/* Hrader - lunghezza testata */
				head = strtol (argv [argc] + 2,(char **) 0,0);
				break;
			case 'r' :
			case 'R' :	/* Record - lunghezza record */
				reclen = ireclen = strtol (argv [argc] + 2,(char **) 0,0);
				break;
			case 'i' :
			case 'I' :	/* Interattivo */
				inter = 1;
				break;
			case 'n' :
			case 'N' :	/* non aggiusta lunghezza a potenza di 8 */
				irlen = rlen = -1;
				break;
			case 'w' :
			case 'W' :	// Word display
				align = 1;
				break;
			case 'd' :
			case 'D' :	// Double Word display
				align = 2;
				break;
			case 'P' :
			case 'p' :
				if (argv [argc][2]) plen = -strtol (argv [argc] + 2,(char **) 0,0);
				else plen = -66;
				if (plen > -1) plen = -66;
				break;
			case '?' :
				printf ("USAGE :\ndump <file 1> [<file 2> ...] [-h] [-s<start>] "
					"[-e<end>] [-c] [-r<len>]\n   [-t<head>] [-n] [-i] [-?]\n\n");
				printf (" <file 1> [<file 2> ...] from 1 to 14 files to dump\n");
				printf (" -h Only hex dump (not ASCII)\n");
				printf (" -s<start> Start of dump in every base (C style)\n");
				printf (" -e<end> End of dump in every base\n");
				printf (" -c Compare - marks differences between files\n");
				printf (" -r<len> Record Len - break lines at this len\n");
				printf (" -t<head> Header Len - records start from this byte\n");
				printf (" -n use max row len (oterwise prefers powers of 2)\n");
				printf (" -i Interactive - browse dump interactively\n");
				printf (" -w Displays data as Ints (words)\n");
				printf (" -d Displays data ad Longs (double words)\n");
				printf (" -p[<nn>] Paginate listing (nn = eventual page "
					"len - default 66)\n");
				printf (" -? Shows this help\n");
				exit (0);
				break;
			default :
				printf ("ERROR Invalid parameter\nuse DUMP -? for help\n");
				exit (1);
		}
	}

	if (argc < 2 || argc > MAXFILS)
	{
		printf ("USAGE :\ndump <file 1> [<file 2> ...] [-h] [-s<start>] "
			"[-e<end>] [-c] [-r<len>]\n   [-t<head>] [-n] [-i] [-w] [-d] [-?]\n");
		exit (1);
	}

	rlen = CalcRLen (&l,argc,hex,rlen,align);
	if (reclen == 0) reclen = rlen;

	if (rlen == 0)
	{
		printf ("Too many files to dump\n");
		exit (2);
	}

	for (i = 1;i < argc;i ++)
	{
		fil [i - 1] = open (argv [i],O_RDONLY | O_BINARY);
		if (fil [i - 1] == EOF)
		{
			printf ("Can't open file %s\n",argv [i]);
			i--;
			while (i--) close (fil [i]);
			exit (1);
		}
	}

	if (inter)
	{
		start = 0;
		cl = -1;

		NewWin (0,0,1,1,80,25,(char *) 0);

		for (i = maxl = 0;i < argc - 1;i ++)
			if (filelength (fil [i]) > maxl)
				maxl = filelength(fil [i]);

		showframe (argc,argv,l);

		do {
			gotoxy (1,2);
			if (start != cl && !bioskey (1))
			{
				PrintLines (22,argc,argv,l,fil,start,0x7fffffff,head,rlen,
					reclen,hex,compare,align);
				cl = start;
			}

			key = Key ();
			gotoxy (70,25);
			printf ("%04X",key);

			switch (key)
			{
				case HOME :	// Home
					start = 0;
					break;
				case ARUP :	// Freccia Su
					if (start <= head)
					{
						rcl = start % rlen;
					}
					else
					{
						rcl = (start - head) % reclen;
						if (rcl == 0) rcl = reclen;
						rcl = rcl % rlen;
					}
					if (rcl <= 0) rcl = rlen;
					start -= rcl;
					if (start < 0) start = 0;
					break;
				case PGUP :	// Pagina Su
					for (i = 0;i < 22 && start;i ++)
					{
						if (start <= head)
						{
							rcl = start % rlen;
						}
						else
						{
							rcl = (start - head) % reclen;
							if (rcl == 0) rcl = reclen;
							rcl = rcl % rlen;
						}
						if (rcl <= 0) rcl = rlen;
						start -= rcl;
						if (start < 0) start = 0;
					}
					break;
				case '-' :		// Record su
					if (start <= head) start = 0;
					else
					{
						rcl = (start - head) % reclen;
						start -= (rcl) ? rcl : reclen;
					}
					break;
				case ARDOW :	// Freccia Giu
					if (start < head)
					{
						rcl = (start + rlen < head) ? rlen : head - start;
					}
					else
					{
						rcl = (start - head) % reclen;
						rcl = (rcl + rlen < reclen) ? rlen : reclen - rcl;
					}
					if (start + rcl < maxl) start += rcl;
					break;
				case PGDOW :	// Pagina Giu
					for (i = 0;i < 22 && start + rcl < maxl;i ++)
					{
						if (start < head)
						{
							rcl = (start + rlen < head) ? rlen : head - start;
						}
						else
						{
							rcl = (start - head) % reclen;
							rcl = (rcl + rlen < reclen) ? rlen : reclen - rcl;
						}
						if (start + rcl < maxl) start += rcl;
					}
					break;
				case '+' :		// Record Giu
					if (start < head) start = head;
					else if (start + reclen < maxl)
						start += reclen - (start - head) % reclen;
					break;
				case END :
					start = maxl - rlen;
					break;
				case 'O' :
				case 'o' :
				case F8 :
					Options (&hex,&compare,&irlen,&rlen,&align,&reclen,
						&ireclen,&start,&l,&head,maxl,argc,argv);

					cl = start - 1;	// Forza ridisegno sfondo
					break;
				case 'C' :
				case 'c' :
				case F9 :	// Calculator
					Calculator ();
					break;
			}

		} while (key != F10 && key != 'Q' && key != 'q');
		gotoxy (1,25);
		clreol ();
	}
	else
	{
		PrintHead (argc,argv,l);
		printf ("\n");

		PrintLines (plen,argc,argv,l,fil,start,end,head,rlen,reclen,hex,
			compare,align);
	}

	for (i = 0;i < argc - 1;i ++)
		close (fil [i]);

	exit (0);
}

void Options (char *hex,char *compare,int *irlen,int *rlen,char *align,
	int *reclen,int *ireclen,long *start,int *l,long *head,long maxl,
		int argc,char **argv)
{
long	tstart,thead;
char    thex,tcompare,talign,trlen;
int 	treclen;
int		i,j,key;
long	cl;

	NewWin (1,1,27,10,54,18,"Options");

	thex = *hex;
	tcompare = *compare;
	trlen = !*irlen;
	talign = *align;
	thead = *head;
	treclen = *ireclen;
	tstart = *start;

	gotoxy (1,1);
	printf ("       Hex Only : %s",(thex) ? "Yes" : "No ");
	gotoxy (1,2);
	printf ("        Compare : %s",(tcompare) ? "Yes" : "No ");
	gotoxy (1,3);
	printf ("2's Power Align : %s",(trlen) ? "Yes" : "No ");
	gotoxy (1,4);
	printf ("        Display : %s",(talign) ? (
		(talign > 1) ? "Long" : "Int ") : "Char");
	gotoxy (1,5);
	printf ("     Header Len : %9ld",thead);
	gotoxy (1,6);
	printf ("  Record Length : %5d",treclen);
	gotoxy (1,7);
	printf (" Start Position : %9ld",tstart);
	gotoxy (1,8);
	printf ("   Which Record ? %9ld",(*start < *head) ? -1 :
		(*start - *head) / *reclen);
	gotoxy (1,9);
	textattr (REVERSE);
	clreol ();
	printf ("  ESC Abandon;F10 Confirm");
	textattr (NORMAL);

	i = 0;	// item 0;
	do {
		switch (i)
		{
			case 0 :	// Hex
				key = Boolean (&thex,19,1);
				break;
			case 1 :	// Compare
				key = Boolean (&tcompare,19,2);
				break;
			case 2 :	// 2's power align
				key = Boolean (&trlen,19,3);
				break;
			case 3 :	// Kind of display
				key = Kind (&talign,19,4);
				break;
			case 4 :	// header
				key = Long (&thead,19,5);
				break;
			case 5 :	// Rec Len
				key = Int (&treclen,19,6);
				break;
			case 6 :	// start
				key = Long (&tstart,19,7);
				if (tstart > maxl)
				{
					tstart = maxl - *rlen;
					gotoxy (19,7);
					cprintf ("%9ld",tstart);
				}
				break;
			case 7 :	// Record
				// Salta se reclen = 0
				if (treclen == 0) break;

				if (tstart < thead) cl = -1;
				else cl = (tstart - thead) / treclen;
				key = Long (&cl,19,8);
				if ((cl != (tstart - thead) / treclen) &&
					(cl != -1 || tstart >thead))
				{
					if (cl < 0) tstart = 0;
					else tstart = thead + cl * treclen;
					if (tstart > maxl)
					{
						cl = (maxl - thead) / treclen;
						tstart = thead + cl * treclen;
					}
					if (maxl < thead)
					{
						cl = -1;
						tstart = 0;
					}
					gotoxy (19,7);
					cprintf ("%9ld",tstart);
					gotoxy (19,8);
					cprintf ("%9ld",cl);
				}
				break;
		}
		switch (key)
		{
			case CR :
			case ARDOW :
				i++;
				if (i >= MAXIN) i = 0;
				break;
			case ARUP :
				i --;
				if (i < 0) i = MAXIN - 1;
				break;
			case F9 :
				Calculator ();
		}

		if (key == F10)
		{
			i = CalcRLen (&j,argc,thex,!trlen,talign);
			if (i < 1)
			{
				Warning ("Too many files for this kind "
					"of display");
				i = 3;
				continue;
			}
			*rlen = i;
			*l = j;
			*hex = thex;
			*compare = tcompare;
			*irlen = !trlen;
			*head = thead;
			*reclen = *ireclen = treclen;
			*align = talign;
			*start = tstart;

			if (*ireclen == 0) *reclen = *rlen;
			SelectWin (0);
			showframe (argc,argv,*l);
			SelectWin (1);
			key = ESC;
		}
	} while (key != ESC);

	clrscr ();
	SelectWin (0);
}

void Warning (char *mess)
{
int s,l,rear;
static char TheVideo [2 * 80 * (13 - 10 + 1) + 4];

	rear = curwin;
	l = strlen (mess);
	if (l > 78) l = 78;
	if (l < 22) l = 22;
	s = (80 - l) / 2;
	gettext (1,10,80,13,TheVideo);
	NewWin (2,1,s,11,s + l,12,(char *) 0);
	gotoxy (1,1);
	cprintf ("%s",mess);
	gotoxy (1,2);
	textattr (REVERSE);
	clreol ();
	gotoxy ((l - 22) / 2,2);
	cprintf ("Press ESC to Continue");
	textattr (NORMAL);

	while (Key () != ESC);

	SelectWin (rear);
	puttext (1,10,80,13,TheVideo);
}

void Calculator (void)
{
int key,rear;
char TheVideo [2 * (79 - 65 +1) * (12 - 2 + 1) + 4];
long op,acc;
char mode,opr;
char edit [20];

	rear = curwin;
	gettext (65,2,79,12,TheVideo);
	NewWin (2,1,66,3,78,11,"Calculator");

	gotoxy (1,4);
	cprintf ("+ - * / & | =");
	gotoxy (1,5);
	cprintf ("Alt C - Clear");
	gotoxy (1,6);
	cprintf ("Alt D - Dec");
	gotoxy (1,7);
	cprintf ("Alt H - Hex");
	gotoxy (1,8);
	cprintf ("Alt O - Oct");
	gotoxy (1,9);
	textattr (REVERSE);
	clreol ();
	printf (" ESC to exit");
	textattr (NORMAL);

	mode = 1;
	acc = 0;
	opr = 'C';
	do {
		gotoxy (1,1);

		switch (mode)
		{
			case 0 :	// Hex
				cprintf ("    %08lX\n\r Hex     ",acc);
				break;
			case 1 :	// Decimal
				cprintf (" %11ld\n\r Decimal",acc);
				break;
			case 2 :	// Octal
				cprintf (" %11lo\n\r Octal  ",acc);
				break;
		}

		edit [0] = 0;
		key = GetInput (edit,11,2,3,calcterm);
		op = strtol (edit,(char **) 0,0);

		switch (key)
		{
			case CR :
			case '=' :
				if (edit [0]) acc = doopr (opr,acc,op);
				break;
			case ALTH :	// Alt H - Hexadecimal
				mode = 0;
				if (edit [0]) acc = doopr (opr,acc,op);
				else key = opr;
				break;
			case ALTD :	// Alt D - Decimal
				mode = 1;
				if (edit [0]) acc = doopr (opr,acc,op);
				else key = opr;
				break;
			case ALTO :	// Alt O - Octal
				mode = 2;
				if (edit [0]) acc = doopr (opr,acc,op);
				else key = opr;
				break;
			case ALTC :	// Alt C - Clear
				acc = 0;
				break;
			case '+' :
			case '-' :
			case '*' :
			case '/' :
			case '&' :
			case '|' :
				if (edit [0]) acc = doopr (opr,acc,op);
				break;
		}
		opr = key;
	} while (key != ESC);

	SelectWin (rear);
	puttext (65,2,79,12,TheVideo);
}



long doopr (char opr,long acc,long op)
{
	switch (opr)
	{
		case '+' :
			acc += op;
			break;
		case '-' :
			acc -= op;
			break;
		case '*' :
			acc *= op;
			break;
		case '/' :
			if (op) acc /= op;
			else acc = 0xffffffff;
			break;
		case '&' :
			acc &= op;
			break;
		case '|' :
			acc |= op;
			break;
		default :
			acc = op;
			break;
	}
	return (acc);
}

void showframe (int argc,char **argv,int l)
{
	textattr (NORMAL);
	clrscr ();
	textattr (REVERSE);
	clreol ();
	PrintHead (argc,argv,l);
	gotoxy (1,25);
	clreol ();
	printf ("F10 Quit;F9 Calculator;F8 Options * Written By S. Salvi");
	textattr (NORMAL);
}

int CalcRLen (int *l,int argc,char hex,int rlen,char align)
{
	/* calcolo lunghezza singolo campo in hex */
	switch (align)
	{
		case 0 :	// char
			//		   Dimensione separatori           numero campi
			*l = (71 - (argc - 1) * ((hex) ? 2 : 3)) / (argc - 1) /
				((hex) ? 3 : 4);
			if (rlen) rlen = *l;	/* parametro -n; usa calcolo massimo */
			else
			{
				/* calcola potenza di 2 subito inferiore */
				for (rlen = 1;rlen * 2 < *l;rlen *= 2);
				/* se il risultato e' troppo diverso, usa valore esatto */
				if (*l - rlen > *l / 4) rlen = *l;
			}
			*l = (rlen * ((hex) ? 3 : 4)) + ((hex) ? 0 : 1);
			break;
		case 1 :	// Int
			//		   Dimensione separatori           numero campi
			*l = (71 - (argc - 1) * ((hex) ? 2 : 3)) / (argc - 1) /
				((hex) ? 5 : 7);
			rlen = *l;
			*l = (rlen * ((hex) ? 5 : 7)) + ((hex) ? 0 : 1);
			rlen *= 2;
			break;
		default :	// Long
			//		   Dimensione separatori           numero campi
			*l = (71 - (argc - 1) * ((hex) ? 2 : 3)) / (argc - 1) /
				((hex) ? 9 : 13);
			rlen = *l;
			*l = (rlen * ((hex) ? 9 : 13)) + ((hex) ? 0 : 1);
			rlen *= 4;
			break;
	}
	return (rlen);
}

int Kind (char *value,int x,int y)
{
int key;
	textattr (REVERSE);
	do {
		gotoxy (x,y);
		cprintf ("%s",(*value) ? ((*value > 1) ? "Long" : "Int ") : "Char");
		key = Key ();
		if (key == ' ')
		{
			(*value)++;
			if (*value > 2) *value = 0;
		}
	} while (key == ' ');
	textattr (NORMAL);
	gotoxy (x,y);
	cprintf ("%s",(*value) ? ((*value > 1) ? "Long" : "Int ") : "Char");
	return (key);
}

int	Boolean (char *value,int x,int y)
{
int key;
	textattr (REVERSE);
	do {
		gotoxy (x,y);
		cprintf ("%s",(*value) ? "Yes" : "No ");
		key = Key ();
		if (key == ' ') *value = !*value;
	} while (key == ' ');
	textattr (NORMAL);
	gotoxy (x,y);
	cprintf ("%s",(*value) ? "Yes" : "No ");
	return (key);
}

int Long (long *value,int x,int y)
{
int key;
char edit [20];

	sprintf (edit,"%ld",*value);
	key = GetInput (edit,9,x,y,stdtetm);
	*value = strtol (edit,(char **) 0,0);
	gotoxy (x,y);
	textattr (NORMAL);
	cprintf ("%9ld",*value);
	return (key);
}

int Int (int *value,int x,int y)
{
int key;
char edit [20];

	sprintf (edit,"%d",*value);
	key = GetInput (edit,5,x,y,stdtetm);
	*value = strtol (edit,(char **) 0,0);
	gotoxy (x,y);
	textattr (NORMAL);
	cprintf ("%5d",*value);
	return (key);
}

int shortcursor;
int tallcursor;

void initcursor(void)
/* Initializes the different cursor types */
{
 struct text_info ti;

 gettextinfo(&ti);
 if (ti.currmode == MONO)
 {
  shortcursor = 0x0A0C;
  tallcursor = 0x060B;
 }
 else
 {
  shortcursor = 0x0607;
  tallcursor = 0x0407;
 }
} /* initcursor */

void setcursor(unsigned int shape)
/* Sets the shape of the cursor */
{
 union REGS reg;

 reg.h.ah = 1;
 reg.x.cx = shape;
 int86(0X10, &reg, &reg);
} /* setcursor */

void changecursor(insmode)
/* Changes the cursor shape based on the current insert mode */
{
	if (!insmode)
		setcursor(tallcursor);
	else
		setcursor(shortcursor);
} /* changecursor */

int strint (int *s,int c)
{
	while (*s && *s != c) s++;
	return (*s);
}

int GetInput (char *s,int max,int st,int y,int *term)
{
int c, len = strlen(s), pos = 0, insert = 0;
int ex;

	initcursor ();

	textattr (REVERSE);
	changecursor(insert);
	ex = 1;
	do {
		gotoxy (st, y);
		cprintf ("%-*s",max, s);
		gotoxy (pos + st, y);
		switch(c = Key ())
		{
			case HOME :
				pos = 0;
				break;
			case END :
				pos = len;
				break;
			case INS :
				insert = !insert;
				changecursor(insert);
				break;
			case LEFT :
				if (pos > 0)
				pos--;
				break;
			case RIGHT :
				if (pos < len)
				pos++;
				break;
			case BS :
				if (pos > 0)
				{
					movmem(&s[pos], &s[pos - 1], len - pos + 1);
					pos--;
					len--;
				}
				break;
			case DEL :
				if (pos < len)
				{
					movmem(&s[pos + 1], &s[pos], len - pos);
					len--;
				}
				break;
			default :
				if (c & 0xff) c = toupper (c);
				if (strint (term,c))
				{
					if (c == ESC) len = 0;
					ex = 0;
					break;
				}
				if (((c >= '0' && c <= '9') || (c >= 'A' && c <= 'F') ||
					c == 'X' || c == '+' || c == '-') && (len < max))
				{
					if (insert)
					{
						memmove(&s[pos + 1], &s[pos], len - pos + 1);
						len++;
					}
					else if (pos >= len)
					len++;
					s[pos++] = c;
				}
				break;
		   } /* switch */
		   s[len] = 0;
	} while (ex);

	changecursor (1);
	textattr (NORMAL);
	return (c);
}

void PrintHead (int argc,char **argv,int l)
{
int i;
char nome [80];

	/* Stampa Testata */
	printf ("         ");	/* Spazio corrispondente alla testata */
	for (i = 1;i < argc;i ++)
	{
		strcpy (nome,argv [i]);
		nome [l] = 0;
		printf ((i < argc - 1) ? " %-*s|" : " %-*s",l,nome);
	}
}

void PrintLines (int lines,int argc,char **argv,int l,int *fil,long start,
	long end,long head,int rlen,int reclen,char hex,char compare,char align)
{
int			f,i,j,n,rl,line;
long		crl;
unsigned char bufs [MAXFILS][24];
int 		lens [MAXFILS];
unsigned 	diffs [24];
char 		fine;

		crl = (head) ? head : reclen;
		setmem (bufs,sizeof (bufs),0);

		if (start || lines > 0)
		{
			for (i = 0;i < argc - 1;i ++)
				lseek (fil [i],start,SEEK_SET);
			if (start < head) crl = head - start;
			else crl = reclen - ((start - head) % reclen);
		}

		line = 0;
		do {
			/* Calcola lunghezza riga per lunghezza record */
			if (crl <= rlen)
			{
				rl = crl;
				crl = reclen;
			}
			else
			{
				rl = rlen;
				crl -= rlen;
			}

			fine = 1;
			for (f = 0;f < argc - 1;f ++)
			{
				lens [f] = read (fil [f],bufs [f],rl);
				if (lens [f] == rl && !eof (fil [f])) fine = 0;
			}

			if (compare)
			{
				/* Per prima cosa guarda se ci sono differenze */
				for (j = i = 0;i < rlen;i ++)
				{
					diffs [i] = 0;
					// Cerca primo file che ha un carttere all'indice i
					for (n = 0;lens [n] <= i && n < argc - 1;n ++);

					// Per i files che seguono (se c'e' carattere) confronta
					for (f = n + 1;f < argc - 1;f ++)
						if (i <= lens [f] && bufs [f][i] != bufs [n][i])
						{
							diffs [i] = 1;
							j = 1;
						}
				}
			}

			printf ("%08lX ",start);
			for (f = 0;f < argc - 1;f ++)
			{
				switch (align)
				{
					case 0 :	// char
						for (i = 0;i < rlen;i ++) if (lines > 0)
						{
							if (compare && diffs [i] && i < lens [f])
								textattr (LOW);
							cprintf ((i < lens [f]) ? " %02X" : "   ",
								bufs [f][i] & 0xff);
							if (compare && diffs [i] && i < lens [f])
								textattr (NORMAL);
						}
						else printf ((i < lens [f]) ? " %02X" : "   ",
							bufs [f][i] & 0xff);
						break;
					case 1 :	// word
						for (i = 0;i < rlen;i += 2) if (lines > 0)
						{
							if (compare && (diffs [i] || diffs [i + 1]) &&
								i < lens [f])
									textattr (LOW);
							cprintf ((i < lens [f]) ? " %04X" : "     ",
								* ((int*) (bufs [f] + i)));
							if (compare && (diffs [i] || diffs [i + 1]) &&
								i < lens [f])
									textattr (NORMAL);
						}
						else printf ((i < lens [f]) ? " %04X" : "     ",
							* ((int*) (bufs [f] + i)));
						break;
					default :	// long
						for (i = 0;i < rlen;i += 4) if (lines > 0)
						{
							if (compare && (diffs [i] || diffs [i + 1] ||
								diffs [i + 2] || diffs [i + 3]) &&
									i < lens [f])
										textattr (LOW);
							cprintf ((i < lens [f]) ? " %08lX" : "         ",
								* ((long*) (bufs [f] + i)));
							if (compare && (diffs [i] || diffs [i + 1] ||
								diffs [i + 1] || diffs [i + 1]) &&
									i < lens [f])
										textattr (NORMAL);
						}
						else printf ((i < lens [f]) ? " %04X" : "     ",
							* ((int*) (bufs [f] + i)));
						break;
				}
				if (!hex)
				{
					printf (" ");
					for (i = 0;i < rlen;i ++) if (lines > 0)
					{
						if (compare && diffs [i] && i < lens [f])
							textattr (LOW);
						cprintf ((i < lens [f]) ? "%c" : " ",
							(bufs [f][i] >= ' ' && bufs [f][i] < 127) ?
								bufs [f][i] : '.');
						if (compare && diffs [i] && i < lens [f])
							textattr (NORMAL);
					}
					else printf ((i < lens [f]) ? "%c" : " ",
							(bufs [f][i] >= ' ' && bufs [f][i] < 127) ?
								bufs [f][i] : '.');
				}
				if (f < argc - 2) printf (" |");
			}
			printf ("\n");

			if (compare && j && lines < 0)
			{	// Solo se c'e' qualche differenza, stampa riga diff.
				line ++;	// conta anche questa linea
				printf ("         ");	/* Spazio corrispondente alla testata */
				for (f = 0;f < argc - 1;f ++)
				{
					switch (align)
					{
						case 0 :	// char
							for (i = 0;i < rlen;i ++)
								printf ((i < lens [f] && diffs [i]) ?
									" ^ " : "   ");
							break;
						case 1 : // Word
							for (i = 0;i < rlen;i += 2)
								printf ((i < lens [f] && (diffs [i] ||
									diffs [i + 1])) ?
										" ^   " : "     ");
							break;
						case 2 : // DoubleWord
							for (i = 0;i < rlen;i += 4)
								printf ((i < lens [f] && (diffs [i] ||
									diffs [i + 1] || diffs [i + 2]||
										diffs [i + 3])) ? " ^     " :
											"       ");
							break;
					}
					if (!hex)
					{
						printf (" ");
						for (i = 0;i < rlen;i ++)
							printf ((i < lens [f] && diffs [i]) ?
								"^" : " ");
					}
					if (f < argc - 2) printf (" |");
				}
				printf ("\n");
			}
			line ++;

			if (lines < -1 && line + lines > -4)
			{
				printf ("\x0C");
				PrintHead (argc,argv,l);
				printf ("\n");
				line = 0;
			}

			start += rl;
		} while ((lines < 0 && !fine && start < end) ||
			(lines > 0 && line <= lines));
}